#include <stdint.h>
#include <string.h>
#include "nordic_common.h"
#include "nrf.h"
#include "app_error.h"
#include "ble.h"
#include "ble_hci.h"
#include "ble_srv_common.h"
#include "ble_advdata.h"
#include "ble_conn_params.h"
#include "softdevice_handler.h"
#include "app_timer.h"
#include "app_button.h"
#include "ble_bls.h"
#include "ble_mcs.h"
#include "bsp.h"
#include "ble_gap.h"
#include "app_pwm.h"
#include "nrf_drv_twi.h"
#include "SEGGER_RTT.h"
#include "nrf_drv_timer.h"
#include "nrf_drv_clock.h"
#include "lis3dh.h"
#include "nrf_delay.h"
#include "debug_ble.h"
#include "nrf_drv_gpiote.h"
#include "BQ27421.h"
#include "app_gpiote.h"
#include "nrf_drv_rtc.h"
#include "pstorage.h"

#define CENTRAL_LINK_COUNT              0                                           /**< Number of central links used by the application. When changing this number remember to adjust the RAM settings*/
#define PERIPHERAL_LINK_COUNT           1                                           /**< Number of peripheral links used by the application. When changing this number remember to adjust the RAM settings*/

#define COMPARE_COUNTERTIME  (1UL)                                        /**< Get Compare event COMPARE_TIME seconds after the counter starts from 0. */

#define DEVICE_NAME                     "Tring"                            				 /**< Name of device. Will be included in the advertising data. */

#define APP_ADV_INTERVAL                120                                          /**< The advertising interval (in units of 0.625 ms; this value corresponds to 40 ms). */
#define APP_ADV_TIMEOUT_IN_SECONDS      BLE_GAP_ADV_TIMEOUT_GENERAL_UNLIMITED       /**< The advertising time-out (in units of seconds). When set to 0, we will never time out. */

#define APP_TIMER_PRESCALER             15                                           /**< Value of the RTC1 PRESCALER register. */
#define APP_TIMER_MAX_TIMERS            (3 + BSP_APP_TIMERS_NUMBER + 6)             /**< Maximum number of simultaneously created timers. */
#define APP_TIMER_OP_QUEUE_SIZE         8                                           /**< Size of timer operation queues. */

#define MIN_CONN_INTERVAL               MSEC_TO_UNITS(100, UNIT_1_25_MS)            /**< Minimum acceptable connection interval (0.5 seconds). */
#define MAX_CONN_INTERVAL               MSEC_TO_UNITS(200, UNIT_1_25_MS)            /**< Maximum acceptable connection interval (1 second). */
#define SLAVE_LATENCY                   0                                           /**< Slave latency. */
#define CONN_SUP_TIMEOUT                MSEC_TO_UNITS(4000, UNIT_10_MS)             /**< Connection supervisory time-out (4 seconds). */
#define FIRST_CONN_PARAMS_UPDATE_DELAY  APP_TIMER_TICKS(20000, APP_TIMER_PRESCALER) /**< Time from initiating event (connect or start of notification) to first time sd_ble_gap_conn_param_update is called (15 seconds). */
#define NEXT_CONN_PARAMS_UPDATE_DELAY   APP_TIMER_TICKS(5000, APP_TIMER_PRESCALER)  /**< Time between each call to sd_ble_gap_conn_param_update after the first call (5 seconds). */
#define MAX_CONN_PARAMS_UPDATE_COUNT    3                                           /**< Number of attempts before giving up the connection parameter negotiation. */
#define BATTERY_LEVEL_MEAS_INTERVAL     APP_TIMER_TICKS(2000, APP_TIMER_PRESCALER) /**< Battery level measurement interval (ticks). */

#define APP_GPIOTE_MAX_USERS            10                                           /**< Maximum number of users of the GPIOTE handler. */

#define DEAD_BEEF                       0xDEADBEEF                                  /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */

uint32_t variable, variable1 , variable2 , variable3 ;
uint32_t set_time[50]={0};


uint32_t epoch_time ;

static uint16_t                         m_conn_handle = BLE_CONN_HANDLE_INVALID;    /**< Handle of the current connection. */
static ble_bls_t                        m_bls;                                      /**< LED Button Service instance. */


uint32_t rtc_time =0;

static ret_code_t BQ27421_ChrgSetup(bool state);
const nrf_drv_rtc_t rtc = NRF_DRV_RTC_INSTANCE(2);

pstorage_handle_t       base_handle;
pstorage_handle_t				block_handle;
static uint8_t pstorage_wait_flag = 0;
static pstorage_block_t pstorage_wait_handle = 0;
static void pstorage_test_store_and_update(void);
void	epoch_pstorage_store(uint32_t source_data);
void	epoch_pstorage_update(uint32_t source_data);
void	epoch_pstorage_load(void);
static uint8_t 													pstorage_flag;
	

static volatile bool ready_flag;            // A flag indicating PWM status.

static void gpiote_init(void) { APP_GPIOTE_INIT(10); }

/**@brief Function for assert macro callback.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyze 
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num    Line number of the failing ASSERT call.
 * @param[in] p_file_name File name of the failing ASSERT call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(DEAD_BEEF, line_num, p_file_name);
}


static void rtc_handler(nrf_drv_rtc_int_type_t int_type)
{
		uint32_t diff;
		char recieved_data[20];
		
	   if (int_type == NRF_DRV_RTC_INT_TICK)
    {					
    		rtc_time ++;
				if(rtc_time %8 == 0) 
				{	
				epoch_time++;
					
			//	sprintf(recieved_data,"RTC:%d\n",epoch_time);
			//	SEGGER_RTT_WriteString(0, recieved_data);
					if(set_time[0] >= epoch_time)
					{
						diff = epoch_time - set_time[0];
							if (diff == 0)
							{
								SEGGER_RTT_WriteString(0, "Alarm\n");
								play_sequence();
								for (int i = 0 ; i < 49 ; i++)
									{
										set_time[i] = set_time[i+1];
									}
									if(event_index > 0)
										event_index--;
									else
										event_index = 0;	
							}
					}
				}
	 			nrf_drv_rtc_counter_clear(&rtc);
					
    }

}

void appRTCInit(void)
{
    uint32_t err_code;
    //Initialize RTC instance
    err_code = nrf_drv_rtc_init(&rtc, NULL, rtc_handler);
    APP_ERROR_CHECK(err_code);
		
		//Enable tick event & interrupt
    nrf_drv_rtc_tick_enable(&rtc,true);

    //Power on RTC instance
    nrf_drv_rtc_enable(&rtc);
}


/**@brief Function for the Timer initialization.
 *
 * @details Initializes the timer module.
 */
static void timers_init(void)
{
		uint32_t err_code;

		// Initialize timer module, making it use the scheduler
    APP_TIMER_INIT(APP_TIMER_PRESCALER, APP_TIMER_OP_QUEUE_SIZE, false);
	
    // Create timers
	
}

/**@brief Function for the GAP initialization.
 *
 * @details This function sets up all the necessary GAP (Generic Access Profile) parameters of the
 *          device including the device name, appearance, and the preferred connection parameters.
 */
static void gap_params_init(void)
{
    uint32_t                err_code;
    ble_gap_conn_params_t   gap_conn_params;
    ble_gap_conn_sec_mode_t sec_mode;

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

    err_code = sd_ble_gap_device_name_set(&sec_mode,
                                          (const uint8_t *)DEVICE_NAME,
                                          strlen(DEVICE_NAME));
    APP_ERROR_CHECK(err_code);

    memset(&gap_conn_params, 0, sizeof(gap_conn_params));

    gap_conn_params.min_conn_interval = MIN_CONN_INTERVAL;
    gap_conn_params.max_conn_interval = MAX_CONN_INTERVAL;
    gap_conn_params.slave_latency     = SLAVE_LATENCY;
    gap_conn_params.conn_sup_timeout  = CONN_SUP_TIMEOUT;

    err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for initializing the Advertising functionality.
 *
 * @details Encodes the required advertising data and passes it to the stack.
 *          Also builds a structure to be passed to the stack when starting advertising.
 */
static void advertising_init(void)
{
    uint32_t      err_code;
    ble_advdata_t advdata;
    ble_advdata_t scanrsp;

    ble_uuid_t adv_uuids[] = {{bls_UUID_LED_SERVICE,0x02}};

    // Build and set advertising data
    memset(&advdata, 0, sizeof(advdata));

    advdata.name_type          = BLE_ADVDATA_FULL_NAME;
    advdata.include_appearance = true;
    advdata.flags              = BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE;


    memset(&scanrsp, 0, sizeof(scanrsp));
    scanrsp.uuids_complete.uuid_cnt = sizeof(adv_uuids) / sizeof(adv_uuids[0]);
    scanrsp.uuids_complete.p_uuids  = adv_uuids;

    err_code = ble_advdata_set(&advdata, &scanrsp);
    APP_ERROR_CHECK(err_code);
}



/**@brief Function for starting application timers.
 */
static void application_timers_start(void)
{
    uint32_t err_code;

 //    Start application timers.
    err_code = app_timer_start(m_battery_timer_id, BATTERY_LEVEL_MEAS_INTERVAL, NULL);
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for handling write events to the LED characteristic.
 *
 * @param[in] p_bls     Instance of LED Button Service to which the write applies.
 * @param[in] led_state Written/desired state of the LED.
 */
static void epoch_write_handler(ble_bls_t * p_lbs, uint8_t * value)
{
		char recieved_data[20];
		uint32_t err_code ;
		uint32_t	time;
		SEGGER_RTT_WriteString(0,"Received Epoch Timer\n");
	
		variable  = value[0];
		variable1 = value[1];
		variable2 = value[2];
		variable3 = value[3];
		
		time = variable3 <<24 | variable2 <<16 | variable1 <<8 | variable ;
		epoch_time = time;
		sprintf(recieved_data,"Data Timer:%x\n",epoch_time);
	  SEGGER_RTT_WriteString(0, recieved_data);

		application_timers_start();
		
}

/**@brief Function for initializing services that will be used by the application.
 */
static void services_init(void)
{
    uint32_t       err_code;
    ble_bls_init_t init;

		SEGGER_RTT_WriteString(0, "Services Initializing..\n");
		init.epoch_write_handler = epoch_write_handler;
	

    err_code = ble_dbg_init(&m_dbg, &dbg_init);
    APP_ERROR_CHECK(err_code);
		
}


/**@brief Function for handling the Connection Parameters Module.
 *
 * @details This function will be called for all events in the Connection Parameters Module that
 *          are passed to the application.
 *
 * @note All this function does is to disconnect. This could have been done by simply
 *       setting the disconnect_on_fail config parameter, but instead we use the event
 *       handler mechanism to demonstrate its use.
 *
 * @param[in] p_evt  Event received from the Connection Parameters Module.
 */
static void on_conn_params_evt(ble_conn_params_evt_t * p_evt)
{
    uint32_t err_code;

    if (p_evt->evt_type == BLE_CONN_PARAMS_EVT_FAILED)
    {
        err_code = sd_ble_gap_disconnect(m_conn_handle, BLE_HCI_CONN_INTERVAL_UNACCEPTABLE);
        APP_ERROR_CHECK(err_code);
    }
}


/**@brief Function for handling a Connection Parameters error.
 *
 * @param[in] nrf_error  Error code containing information about what went wrong.
 */
static void conn_params_error_handler(uint32_t nrf_error)
{
    APP_ERROR_HANDLER(nrf_error);
}


/**@brief Function for initializing the Connection Parameters module.
 */
static void conn_params_init(void)
{
    uint32_t               err_code;
    ble_conn_params_init_t cp_init;

    memset(&cp_init, 0, sizeof(cp_init));

    cp_init.p_conn_params                  = NULL;
    cp_init.first_conn_params_update_delay = FIRST_CONN_PARAMS_UPDATE_DELAY;
    cp_init.next_conn_params_update_delay  = NEXT_CONN_PARAMS_UPDATE_DELAY;
    cp_init.max_conn_params_update_count   = MAX_CONN_PARAMS_UPDATE_COUNT;
    cp_init.start_on_notify_cccd_handle    = BLE_GATT_HANDLE_INVALID;
    cp_init.disconnect_on_fail             = false;
    cp_init.evt_handler                    = on_conn_params_evt;
    cp_init.error_handler                  = conn_params_error_handler;

    err_code = ble_conn_params_init(&cp_init);
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for starting advertising.
 */
static void advertising_start(void)
{
    uint32_t             err_code;
    ble_gap_adv_params_t adv_params;

    // Start advertising
    memset(&adv_params, 0, sizeof(adv_params));

    adv_params.type        = BLE_GAP_ADV_TYPE_ADV_IND;
    adv_params.p_peer_addr = NULL;
    adv_params.fp          = BLE_GAP_ADV_FP_ANY;
    adv_params.interval    = APP_ADV_INTERVAL;
    adv_params.timeout     = APP_ADV_TIMEOUT_IN_SECONDS;

    err_code = sd_ble_gap_adv_start(&adv_params);
    APP_ERROR_CHECK(err_code);
	
		
}


/**@brief Function for the Power Manager.
 */
static void power_manage(void)
{
    uint32_t err_code = sd_app_evt_wait();
    APP_ERROR_CHECK(err_code);
		
}


static void example_cb_handler(pstorage_handle_t  * handle,
															 uint8_t              op_code,
                               uint32_t             result,
                               uint8_t            * p_data,
                               uint32_t             data_len)
{
	//	if(handle->block_id == pstorage_wait_handle) { pstorage_wait_flag = 0; }  //If we are waiting for this callback, clear the wait flag.
	
		switch(op_code)
		{
			case PSTORAGE_LOAD_OP_CODE:
				 if (result == NRF_SUCCESS)
				 {
					 pstorage_wait_flag = 0;	
						 SEGGER_RTT_WriteString(0,"pstorage LOAD callback received \r\n");
				 }
				 else
				 {
						 SEGGER_RTT_WriteString(0,"pstorage LOAD ERROR callback received \r\n");
					}
				 break;
			case PSTORAGE_STORE_OP_CODE:
				 if (result == NRF_SUCCESS)
				 {
						pstorage_wait_flag = 0;                                    //Set the wait flag. 
						 SEGGER_RTT_WriteString(0,"pstorage STORE callback received \r\n");
					}
				 else
				 {
					   SEGGER_RTT_WriteString(0,"pstorage STORE ERROR callback received \r\n");
					}
				 break;				 
			case PSTORAGE_UPDATE_OP_CODE:
				 if (result == NRF_SUCCESS)
				 {
					 pstorage_wait_flag = 0;
						 SEGGER_RTT_WriteString(0,"pstorage UPDATE callback received \r\n");
					}
				 else
				 {
						 SEGGER_RTT_WriteString(0,"pstorage UPDATE ERROR callback received \r\n");
					}
				 break;
			case PSTORAGE_CLEAR_OP_CODE:
				 if (result == NRF_SUCCESS)
				 {
							pstorage_wait_flag = 0;
						 SEGGER_RTT_WriteString(0,"pstorage CLEAR callback received \r\n");
				 }
				 else
				 {
					   SEGGER_RTT_WriteString(0,"pstorage CLEAR ERROR callback received \r\n");
					}
				 break;
		}			
}

static void pstorage_test_store_and_update(void)
{
		uint32_t                retval;
		uint8_t								  i;
		uint32_t 								count;
		
		pstorage_module_param_t param;
		
		uint8_t 								dest_data_0[4];	
	
		char										recieved_data[20];
	
	
		retval = pstorage_init();
		if(retval != NRF_SUCCESS)
		{
				SEGGER_RTT_WriteString(0,"Error initializing Pstorage\n");
		}
	     
		param.block_size  = 32;                   
		param.block_count = 1;                   
		param.cb          = example_cb_handler;   //Set the pstorage callback handler
			
		SEGGER_RTT_WriteString(0,"\r\n\r\npstorage initializing ... \r\n");
		retval = pstorage_register(&param, &base_handle);
		if (retval == NRF_SUCCESS)
		{
				SEGGER_RTT_WriteString(0,"Pstorage register Success\n");
		}
		
				
}

void epoch_pstorage_store(uint32_t source_data)
{
	uint8_t 								epoch_data[4] __attribute__((aligned(4)));
	uint32_t                retval;
	
		epoch_data[3] = (source_data >> 24 ) & 0xFF;
		epoch_data[2] = (source_data >> 16 ) & 0xFF;
		epoch_data[1] = (source_data >> 8) & 0xFF;
		epoch_data[0] = source_data & 0xFF;
		//Get block identifiers
		pstorage_block_identifier_get(&base_handle, 0, &block_handle);
		
		//Store data to blocks. Wait for the last store operation to finish before reading out the data.
		SEGGER_RTT_WriteString(0,"pstorage store into block 0 \r\n");
		pstorage_store(&block_handle, epoch_data, 32, 0);     //Write to flash, only one block is allowed for each pstorage_store command
		nrf_delay_ms(50);
}

	void epoch_pstorage_load()
	{
			uint8_t 								dest_data_0[4];	
			uint32_t                retval;
			uint8_t								  i;
			uint32_t 								count;
			
			char 										recieved_data[20];					
	
		//Get block identifiers
			pstorage_block_identifier_get(&base_handle, 0, &block_handle);
			SEGGER_RTT_WriteString(0,"pstorage load blocks 0 \r\n");
			pstorage_load(dest_data_0, &block_handle, 32, 0);				 //Read from flash, only one block is allowed for each pstorage_load comman		
			for( i= 0; i<4 ;i++)
			{
			sprintf(recieved_data," Load Data:%x\n",dest_data_0[i]);
			SEGGER_RTT_WriteString(0, recieved_data);
			}

		//Get block identifiers
			pstorage_block_identifier_get(&base_handle, 0, &block_handle);
			pstorage_clear(&block_handle, 32);
	}

 void epoch_pstorage_update(uint32_t source_data)
 {
           uint8_t  source[4] = {(source_data >> 24) & 0xFF, 
																 (source_data >> 16) & 0xFF, 
																 (source_data >> 8) & 0xFF, 
																 (source_data >> 0) & 0xFF};

           pstorage_block_identifier_get(&base_handle, 0, &block_handle);
           pstorage_wait_handle = block_handle.block_id;
           pstorage_wait_flag = 1;
           pstorage_update(&block_handle, source, 32, 0);
           while(pstorage_wait_flag) { power_manage(); }
 }
 
 
/**@brief Function for handling the Application's BLE stack events.
 *
 * @param[in] p_ble_evt  Bluetooth stack event.
 */
static void on_ble_evt(ble_evt_t * p_ble_evt)
{
    uint32_t err_code;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
						SEGGER_RTT_WriteString(0, "Connected\n");
            m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
						task =1;
            err_code = app_button_enable();
            APP_ERROR_CHECK(err_code);
						err_code = BQ27421_ChrgSetup(0);
             break;

        case BLE_GAP_EVT_DISCONNECTED:
						epoch_pstorage_store(epoch_time);
						SEGGER_RTT_WriteString(0, "Disconnected\n");
						m_conn_handle = BLE_CONN_HANDLE_INVALID;				
					  err_code = app_button_disable();
            APP_ERROR_CHECK(err_code);
            advertising_start();
            break;

        case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
            // Pairing not supported
            err_code = sd_ble_gap_sec_params_reply(m_conn_handle,
                                                   BLE_GAP_SEC_STATUS_PAIRING_NOT_SUPP,
                                                   NULL,
                                                   NULL);
            APP_ERROR_CHECK(err_code);
            break;

        case BLE_GATTS_EVT_SYS_ATTR_MISSING:
            // No system attributes have been stored.
            err_code = sd_ble_gatts_sys_attr_set(m_conn_handle, NULL, 0, 0);
            APP_ERROR_CHECK(err_code);
            break;

        default:
            // No implementation needed.
            break;
    }
}


/**@brief Function for dispatching a BLE stack event to all modules with a BLE stack event handler.
 *
 * @details This function is called from the scheduler in the main loop after a BLE stack
 *          event has been received.
 *
 * @param[in] p_ble_evt  Bluetooth stack event.
 */
static void ble_evt_dispatch(ble_evt_t * p_ble_evt)
{
	//	SEGGER_RTT_WriteString(0, "dispatch...\n");
    on_ble_evt(p_ble_evt);
    ble_conn_params_on_ble_evt(p_ble_evt);
    ble_bls_on_ble_evt(&m_bls, p_ble_evt);
	  ble_mcs_on_ble_evt(&m_mcs, p_ble_evt);
	  ble_dbg_on_ble_evt(&m_dbg, p_ble_evt);
	
}

static void sys_evt_dispatch(uint32_t sys_evt)
{
		pstorage_sys_event_handler(sys_evt);
}

/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void)
{
	
    uint32_t err_code;
	    
    nrf_clock_lf_cfg_t clock_lf_cfg = NRF_CLOCK_LFCLKSRC;
   
    // Initialize the SoftDevice handler module.
    SOFTDEVICE_HANDLER_INIT(&clock_lf_cfg, NULL);
    ble_enable_params_t ble_enable_params;
		ble_enable_params.common_enable_params.vs_uuid_count = 2;
		SEGGER_RTT_WriteString(0,"Stack..");
    err_code = softdevice_enable_get_default_config(CENTRAL_LINK_COUNT,
                                                    PERIPHERAL_LINK_COUNT,
                                                    &ble_enable_params);
    APP_ERROR_CHECK(err_code);

    
    //Check the ram settings against the used number of links
    CHECK_RAM_START_ADDR(CENTRAL_LINK_COUNT,PERIPHERAL_LINK_COUNT);
    
		
    // Enable BLE stack.
    err_code = softdevice_enable(&ble_enable_params);
    APP_ERROR_CHECK(err_code);

    ble_gap_addr_t addr;

    err_code = sd_ble_gap_address_get(&addr);
    APP_ERROR_CHECK(err_code);
    err_code = sd_ble_gap_address_set(BLE_GAP_ADDR_CYCLE_MODE_NONE, &addr);
    APP_ERROR_CHECK(err_code);

    // Subscribe for BLE events.
    err_code = softdevice_ble_evt_handler_set(ble_evt_dispatch);
    APP_ERROR_CHECK(err_code);
		
		    // Register with the SoftDevice handler module for BLE events.
    err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
    APP_ERROR_CHECK(err_code);

}



/**@brief Function for application main entry.
 */
int main(void)
{
		ret_code_t err_code ;
		char recieved_data[20];
		
	SEGGER_RTT_WriteString(0, "Initialization...\n");
	// Initialize.
    timers_init();
    ble_stack_init();
	
	// Pstorage initialize
	pstorage_init();
		
	gap_params_init();
    services_init();
    advertising_init();
    conn_params_init();
	gpiote_init();
			
	SEGGER_RTT_WriteString(0, "Advertising starts..\n");
    advertising_start();
		
	appRTCInit();
	err_code = app_timer_start(m_output_manager_id, APP_TIMER_TICKS(500, APP_TIMER_PRESCALER), NULL);
    APP_ERROR_CHECK(err_code);
    
	// pstorage load
	epoch_pstorage_load();

	// Enter main loop.
    for (;;)
    {
        power_manage();
    }
}

/**
 * @}
 */
